// 物理内存分配
#include "Type_defs.h"
#include "memlayout.h"
#include "riscv.h"
#include "stdio.h"
#include "string.h"
#include "memory.h"
#include "spinlock.h"

extern char end[];				// 定义在linker.ld中，内核结束后的第一个地址
								// 已经页对齐了
struct run{
    struct run *next;
};

struct{
	struct spinlock mem_lock;	//锁
    struct run *freelist;		//空闲的物理块
}kmem;

/*
 * @brief 将物理页释放
 * @param pa 物理地址，必须页对齐
*/
void memfree(uint64_t pa){
	struct run *r;
	// 安全检查 地址未对齐  || 地址越界
	if(pa % PGSIZE != 0 || pa < (uint64_t)end || pa >= PYHSTOP){
		panic("memfree\n");
	}
	// 为了防止悬空引用索引到旧数据，这里将内存填满1，若有悬空引用，会直接触发中断
	memset((void *)pa, 1, PGSIZE);
	r = (struct run*)pa;
	
	acquire_lock(&kmem.mem_lock);
	r->next = kmem.freelist;
	kmem.freelist = r;
	release_lock(&kmem.mem_lock);
}

/*
 * @brief 初始化物理内存分配
*/
void init_mem(){
	// 初始化锁 
	init_lock(&kmem.mem_lock, "memorylock");
	kmem.freelist = NULL;
	// 将end到PYHSTOP的空间加入到空闲页列表
	uint64_t p = PGROUNDDOWN((uint64_t)end);
	for(; p+PGSIZE <= PYHSTOP; p += PGSIZE){
		//printf("%x\n", p);
		memfree(p);
	}
	printf("memory init finished\n");
}

/*
 * @brief 分配实页,返回其物理地址
 * @retval return pyshical address of page or 0 if failed
*/
void* alloc_page(){
	struct run *r;
	acquire_lock(&kmem.mem_lock);
	r = kmem.freelist;
	if(r == NULL){
		printf("error: no free memory!\n");
		// panic()  mark
		return 0;
	}
	kmem.freelist = r->next;
	release_lock(&kmem.mem_lock);
	return (void*)r;
}

/*
 * @brief 分配连续的n页内存
*/
void* alloc_npages(int32_t n){
	if(n < 1)	return NULL;
	struct run *last, *start;
	int i;
	acquire_lock(&kmem.mem_lock);
	start = last = kmem.freelist;
	for(i = 0; i < n; i++){
		if(last == NULL){
			printf("error: no free memory!\n");
			return 0;
		}
		last = last->next;
	}
	kmem.freelist = last;
	release_lock(&kmem.mem_lock);
	return (void*)start;
}
