#include "process.h"
#include "spinlock.h"
#include "Type_defs.h"
#include "defs.h"
#include "register.h"
#include "stdio.h"
#include "hart.h"
#include "memory.h"
#include "memlayout.h"
#include "string.h"
#include "schedule.h"

extern uint64_t KernelStack_start;          // 定义在start.S
extern pgtable_t kernel_pagetable;          // 定义在vm.c
extern char trampoline[];                   // 定义在linker.ld trampoline.S
void forkrets(struct trapframe *tf);        // 定义在kernel_trapentry.S 
void switch_to(struct context *old, struct context *new);   // 定义在swich.S
void kernel_thread_entry();

struct{
    struct PCB *next;
    struct spinlock lock;
}next_PCB;                                  // 下一个要运行的进程

struct PCB PCBs[MAX_PROC_NUM];
struct spinlock pid_lock;
pid_t next_pid = 1;
// struct PCB *init_process = NULL;
pgtable_t alloc_pgtable();
static void free_PCB(struct PCB *p);
static inline void free_pgtable(pgtable_t pagetable);
static void forkret(void);
pid_t do_fork(uint32_t clone_flags, uint64_t user_stack, struct trapframe *tf);
static void copy_thread(struct PCB *p, uint64_t esp, struct trapframe *tf);

/*
 * @brief 将所有pcb初始化
*/
void init_proc(){
    struct PCB *p;
    for(int i = 0; i < MAX_PROC_NUM; i++){
        p = &PCBs[i];
        init_lock(&(p->lock), "pcb_lock");
        PCBs[i].state = UNUSED;
    }
    init_lock(&pid_lock, "pid_lock");
    init_lock(&(next_PCB.lock), "next_PCB");
    next_PCB.next = NULL;
}

/*
 * @brief 为当前hart设置idle进程，pid = 0
 * 需要注意的是该进程占用的pcb是PCBs[hartid]
 * 以后需要调用当前hart的idle进程，直接将PCBs[hartid]执行即可
*/
void init_hart_idle(uint64_t hartid){
    struct PCB *p = &PCBs[hartid];
    p->pid = 0;
    p->priority = -1;   //mark
    p->name = "idle";
    p->parent = NULL;
    p->pagetable = (pte_t*)r_satp();      // 这个感觉好像不需要
    p->state = RUNNABLE;
    p->kstack = (KernelStack_start + hartid*KERNEL_STACK_SIZE);   //guard page以后再加吧 mark
    
    // 将pcb信息写入到当前hart上
    push_interrupt();
    myhart()->proc = p;
    pop_interrupt();

    // 修改sp寄存器等 mark process_running();

    printf("hartid %d :init_hart_idle\n", hartid);
}

/*
 * @brief 从PCBs[]里分配一个未使用的PCB，并初始化
 * @retval 如果失败，返回NULL，否则返回PCB指针
 * @note 前KERNEL_MAX_NUM个PCB始终被idle占用
 * @note 
*/
struct PCB* alloc_PCB(void){
    struct PCB *p;
    for(int i = KERNEL_MAX_NUM; i < MAX_PROC_NUM; i++){
        p = &PCBs[i];
        if(p->state == UNUSED){
            if(init_PCB(p) == false)    return NULL;
            return p;
        }
    }
    return NULL;
}

bool init_PCB(struct PCB* p ){
    acquire_lock(&p->lock);
    p->state = SLEEPING;
    p->pid = alloc_pid();
    p->parent = NULL;
    p->runtime = 0;
    p->size_of_memory = 0;
    p->pagetable = alloc_pgtable();
    if(p->pagetable == 0){
        free_PCB(p);
        return false;
    }
    p->kstack = KSTACK((int) (p - PCBs));
    p->name = NULL;

    p->context.ra = (uint64_t)forkret;
    p->context.sp = p->kstack + PGSIZE;
    // 映射内核栈， 同时还有个guard page
    char *pa = alloc_page();
    if(pa == 0)
      panic("alloc_page");
    map_pages(kernel_pagetable, p->kstack, PGSIZE, (uint64_t)pa, PTE_R | PTE_W);
    release_lock(&p->lock);
    return true;
}

/*
 * @brief 分配一个pid
 * @retval pid, 范围[1,100]
*/
pid_t alloc_pid(){
    pid_t pid;
    acquire_lock(&pid_lock);
    pid = next_pid;
    next_pid = (next_pid % 100) + 1;
    release_lock(&pid_lock);
    return pid;
}

struct PCB* myPCB(){
    push_interrupt();
    struct PCB *p = myhart()->proc;
    pop_interrupt();
    return p;
}
/*
 * @brief 为进程分配一张页表
*/
pgtable_t alloc_pgtable(){
    pgtable_t pg;
    pg = (pgtable_t)alloc_page();
    if(pg == 0) return 0;
    memset(pg, 0, PGSIZE);

    if(map_pages(pg, TRAMPOLINE, PGSIZE, (uint64_t)trampoline, PTE_R | PTE_X) == false){
        return 0;
    }
    return pg;
}

static void free_PCB(struct PCB *p){
    memset(&(p->trapframe), 0, PGSIZE);
    p->state = UNUSED;
    free_pgtable(p->pagetable);
}


static inline void free_pgtable(pgtable_t pagetable){
    memfree((uint64_t)pagetable);
}

/*
 * @brief 创建一个内核进程(线程)
 * @param fn 线程要执行的函数
 * @param arg 传递给fn的参数
 * @param clone_flags 共享内存还是独占内存 mark
*/
pid_t kernel_thread(int (*fn)(void *), void *arg, uint32_t clone_flags){
    struct trapframe tpframe;
    memset(&tpframe, 0, sizeof(struct trapframe));

    tpframe.s0 = (uint64_t)fn;
    tpframe.s1 = (uint64_t)arg;
    /* @note kernel_thread_entry是一个小跳板，他将arg(s1 register)移到a0，然后跳转到s0，
    *  即s0所指的函数。执行完后，跳转到do_exit  mark
    */
    tpframe.epc = (uint64_t)kernel_thread_entry;
    // return do_fork(clone_flags | CLONE_VM, 0, &tpframe);
    return do_fork(clone_flags, 0, &tpframe);
}

/*
 * @brief 
 * @param stack 当stack=0时fork内核线程，不分配用户栈，sp始终指向内核栈；
 *              否则stack是父进程的用户栈地址，fork一个用户级线程，我们这里还没实现用户级线程 mark
 * @retval return pid or -1 if failied
 * @details 分配PCB
 *          根据 clone_flags确定是复制还是共享内存 mark
 *          设置进程的中断帧和上下文
 *          将返回值设为线程id
*/
pid_t do_fork(uint32_t clone_flags, uint64_t user_stack, struct trapframe *tf){
    struct PCB *p = alloc_PCB();
    if(p == NULL){
        return -1;
    }
    // clone_flags 内存复制还没实现 mark
    // 参见xv6 的uvmcopy
    p->pagetable = kernel_pagetable; //mark
    copy_thread(p, user_stack, tf);
    acquire_lock(&p->lock);
    p->state = RUNNABLE;
    release_lock(&p->lock);
    return p->pid;
}

static void copy_thread(struct PCB *p, uint64_t esp, struct trapframe *tf){
    p->trapframe = *tf;
    p->trapframe.a0 = 0;        // fork的子进程是返回0
    /*
     * esp==0,内核线程，sp指向内核栈； 
     * esp!=0,sp指向父进程的用户栈
    */
    p->trapframe.sp = (esp == 0) ? p->kstack + PGSIZE : esp;   
    p->context.ra = (uint64_t)forkret;
    p->context.sp = p->kstack + PGSIZE;
}

/*
 * @brief 新线程/进程的第一个内核入口点
 * @details 在do_fork()->copy_thread() 中被写入到pcb->context.ra
*/
static void forkret(void){  
    forkrets(&(myPCB()->trapframe));
}

// 完成系统调用exit() mark 
int32_t do_exit(int32_t error_code){
    panic("process exit!!.\n");
    return 0;
}
// init 进程 pid = 1
int32_t init_process(void *arg) {
    printf("this initproc, pid = %d, name = \"%s\"\n", myPCB()->pid, myPCB()->name);
    printf("To U: \"%s\".\n", (const char *)arg);
    return 0;
}

void init_proc_test(){
    int pid = kernel_thread(init_process, "Hello world!!", 0);
    struct PCB *p = pid2PCB(pid);
    p ->name = "init_proc";
    printf("pid:%d; pro_name:%s;\n",p->pid, p->name);
}

/*
 * @brief 通过pid找到PCB
 * @retval 失败返回NULL
*/
struct PCB* pid2PCB(pid_t pid){
    for(int i = 0; i < MAX_PROC_NUM; i++){
        if(PCBs[i].pid == pid)  return &PCBs[i];
    }
    return NULL;
}

/*
 * @brief idle进程最后执行的函数，陷入死循环，不断调度 mark
*/
void idle(){
    while(1){
        schedule();
    }
}

/*
 * @brief 运行进程p
*/
void proc_run(struct PCB *p){
    /*
    acquire_lock(&next_PCB.lock);
    next_PCB.next = p;
    release_lock(&next_PCB.lock);
    */

    struct PCB *old = myPCB();
    push_interrupt();
    myhart()->proc = p;
    pop_interrupt();

    switch_to(&(old->context), &(p->context));
}