#ifndef _RIGISTER_
#define _RIGISTER_
#include "Type_defs.h"
#include "memlayout.h"
#include "riscv.h"
/*
 * @brief get the hartid
 * @retval the hartid
*/
static inline uint64_t r_mhartid(void){
  uint64_t ret;
  asm volatile(
    "csrr %0, mhartid" 
    : "=r" (ret) 
  );
  return ret;
}

/*
 * @brief 读取tp寄存器
 * @note tp寄存器一般存放hartid
*/
static inline uint64_t r_tp(){
    uint64_t ret;
    asm volatile("mv %0, tp" : "=r" (ret) );
    return ret;
}

/*
 * @brief 写tp寄存器
 * @note tp寄存器一般存放hartid
*/
static inline void  w_tp(uint64_t w){
    asm volatile("mv tp, %0" : : "r" (w));
}

//与trap有关

/*
 * @brief 读取sstatus
 * @retval sstatus寄存器的值
*/
static inline uint64_t r_sstatus(){
    uint64_t ret;
    asm volatile(
        "csrr %0, sstatus" 
        : "=r" (ret) 
    );
    return ret;
}

/*
 * @brief 写sstatus寄存器
 * @param 写入值
*/
static inline void w_sstatus(uint64_t w){
    asm volatile(
        "csrw sstatus, %0" 
        : 
        : "r" (w)
    );
}

/*
 * @brief 读取sepc寄存器  sepc指向发生异常的指令
 * @retval sepc寄存器的值
*/
static inline uint64_t r_sepc(){
    uint64_t ret;
    asm volatile(
        "csrr %0, sepc" 
        : "=r" (ret) 
    );
    return ret;
}

/*
 * @brief 读取scause寄存器 它指示发生异常的种类
 * @note scause寄存器值的具体含义参见 riscv-privileged-20190608-1.pdf
 * P62
 * @retval scause的值
*/
static inline uint64_t r_scause(){
    uint64_t ret;
    asm volatile(
        "csrr %0, scause" 
        : "=r" (ret) 
    );
    return ret;
}

/*
 * @brief 写stvec寄存器
 * @note stvec寄存器保存了trap入口向量的地址，即kernel_trapentry.S
*/
static inline void w_stvec(uint64_t w){
    asm volatile(
        "csrw stvec, %0" 
        : 
        : "r" (w)
    );
}

/*
 * @brief 读stvec寄存器
 * @note stvec寄存器保存了trap入口向量的地址，即kernel_trapentry.S
*/
static inline uint64_t r_stvec(){
    uint64_t ret;
    asm volatile(
        "csrr %0, stvec" 
        : "=r" (ret) 
    );
    return ret;
}

// SIE寄存器， 不是SSTATUS寄存器的sie位
static inline void w_sie(uint64_t w){
    asm volatile(
        "csrw sie, %0" 
        : 
        : "r" (w)
    );
}

static inline uint64_t r_sie(){
    uint64_t ret;
    asm volatile(
        "csrr %0, sie" 
        : "=r" (ret) 
    );
    return ret;
}

// SATP寄存器
static inline void w_satp(uint64_t w){
    asm volatile(
        "csrw satp, %0" 
        : 
        : "r" (w)
    );
}

static inline uint64_t r_satp(){
    uint64_t ret;
    asm volatile(
        "csrr %0, satp" 
        : "=r" (ret) 
    );
  return ret;
}

// 中断 
/*
 * @brief 开启s-mode中断
*/
static inline void s_intr_enable(void){ 
    w_sstatus(r_sstatus() | SSTATUS_SIE);
}

/*
 * @brief 关闭s-mode中断 
*/
static inline void s_intr_disable(void){ 
    w_sstatus(r_sstatus() & ~SSTATUS_SIE);
}

/*
 * @brief 获取s-mode中断使能的情况
 * @retval 1 for enable or 0 for disable
*/
static inline uint64_t s_intr_get(void){
    uint64_t ret;
    ret = (r_sstatus() & SSTATUS_SIE) ;
    return ret != 0;
}

#endif