#include "spinlock.h"
#include "register.h"
#include "hart.h"
#include "stdio.h"

void push_interrupt();
void pop_interrupt();

/*
 * @brief 初始化一个自旋锁
*/
void init_lock(struct spinlock *lock, char *name){
    lock->locked = 0;
    lock->name = name;
    lock->hart = 0;
}

/*
 * @brief 当前hart是否持有该锁
 * @retval false and true if holding
*/
bool is_holding_lock(struct spinlock *lock){

    return (lock->locked && lock->hart == myhart());
}

/*
 * @brief 加锁；被占用后不断test and set
 * @note 不关中断的话可能会导致死锁，但不能直接使用关中断，一个hart里可能有多个
 *          函数关中断，当a函数关中断后，调用b函数，b函数再关中断、开中断，这个时候a函数的关中断就会被打开
 *          这里使用push_interrupt()实现嵌套的关中断
*/
void acquire_lock(struct spinlock *lock){
    
    push_interrupt();
    // 判断hart是否持有锁,如果该hart已经持有该锁，则报错
    if(is_holding_lock(lock)){
        panic("reacquire");
    }
    //test and set
    while(__sync_lock_test_and_set(&lock->locked, 1) != 0) ;
    __sync_synchronize();
    lock->hart = myhart();
}

void release_lock(struct spinlock *lock){
    //holding
    if(!is_holding_lock(lock)){
        panic("release");
    }
    lock->hart = 0;
    __sync_synchronize();
    __sync_lock_release(&lock->locked);
    pop_interrupt();
}

void push_interrupt(){
    struct hart *hart = myhart();
    uint64_t pre_interrupt = s_intr_get();
    s_intr_disable();
    if(hart->interrupt_depth == 0){
        hart->pre_interrupt = pre_interrupt;
    }
    (hart->interrupt_depth)++;
}

void pop_interrupt(){
    struct hart *hart = myhart();
    if(s_intr_get()){
        panic("pop_interrpt 1");
    }
    if(hart->interrupt_depth < 1){
        panic("pop_interrpt 2");
    }
    hart->interrupt_depth -= 1;
    if(hart->interrupt_depth == 0 && hart->pre_interrupt == true){
        s_intr_enable();
    }
}