#include "trap.h"
#include "riscv.h"
#include "stdio.h"
#include "sbi.h"
#include "register.h"
#include "hart.h"

extern void kernel_trap_entry();
static volatile uint64_t ticks;

void init_trap(){
    //参考ucore的做法 https://1790865014.gitbook.io/ucore-step-by-step/intro-2/2_trap
    //约定：若中断前处于S态，sscratch为0
    //若中断前处于U态，sscratch存储内核栈地址

    //不参考了
    s_intr_enable();    // 开启中断使能
    w_stvec((uint64_t)kernel_trap_entry);
    printf("init_trap success!\n");
}
/*
 * @brief  判断是中断还是异常引发的，然后交给各自的处理函数
*/
void kerneltrap(void){
    uint64_t cause =  r_scause() >> (XLEN-1);
    // RISC-V-Reader-Chinese-v2p1.pdf p100
    if(cause == 1){
        //中断
        interrupt_handdle();
    }else if(cause == 0){
        //异常
        exception_handler();
    }else{
        panic("kerneltrap ERROR");
    }
}

/*
 * @brief 中断处理
 * @note mark 待完善
*/
void interrupt_handdle(){
    uint64_t excp_code = (r_scause() << 1) >> 1;    //去掉最高位
    switch(excp_code){
        case U_SOFTWARE_INTERRUPT:
            printf("User software interrupt\n");
            break;
        case S_SOFTWARE_INTERRUPT:
            printf("Supervisor software interrupt\n");
            break;
        case U_TIMER_INTERRUPT:
            printf("User timer interrupt\n");
            break;
        case S_TIMER_INTERRUPT:
            //printf("Supervisor timer interrupt\n");
            clock_set_next_interrupt();
            if( (++ticks) % 500 == 0 ){
                printf("hartid: %d  500 ticks\n",myhartid());
            }
            break;
        case U_EX_INTERRUPT:
            printf("User external interrupt\n");
            break;
        case S_EX_INTERRUPT:
            printf("Supervisor external interrupt\n");
            break;
        default:
            panic("interrupt_handdle");
    }
}

/*
 * @brief 异常处理
 * @note mark 待完善
*/
void exception_handler(){
    uint64_t excp_code = r_scause();
    switch (excp_code){
        case Instruction_Address_Misaligned:
            printf("Instruction Address Misaligned\n");
            break;
        case Instruction_access_fault:
            printf("Instruction access fault\n");
            //while(1){};
            break;
        case Illegal_instruction:
            printf("Illegal instruction\n");
            break;
        case Breakpoint:
            printf("Breakpoint\n");
            break;
        case Load_address_misaligned:
            printf("Load address misaligned\n");
            break;
        case Load_access_fault:
            printf("Load access fault\n");
            break;
        case Store_or_AMO_address_misaligned:
            printf("Store/AMO address misaligned\n");
            break;
        case Store_or_AMO_access_fault:
            printf("Store/AMO access fault\n");
            break;
        case Environment_call_from_Umode:
            // mark 实现系统调用所需要的
            printf("Environment call from Umode\n");
            break;
        case Environment_call_from_Smode:
            printf("Environment call from Smode\n");
            break;
        case Instruction_page_fault:
            printf("Instruction page fault\n");
            panic("Instruction page fault");
            break;
        case Load_page_fault:
            // mark 加载页错误
            printf("Load page fault\n");
            panic("1");
            break;
        case Store_or_AMO_page_fault:
            printf("Store/AMO page fault\n");
            break;
        default:
            panic("interrupt_handdle");
    }
}

//时钟中断相关
//参考了ucore的部分实现思路 https://1790865014.gitbook.io/ucore-step-by-step/intro-2/3_clock
static inline uint64_t get_time(void){
    uint64_t n;
    asm volatile(
        "rdtime %0"
        : "=r"(n)
    );
    return n;
}

void clock_init(void){
    //注意，这里是sie寄存器，不是sstatus寄存器的sie段
    w_sie(r_sie() | SIE_STIE);
    clock_set_next_interrupt();
    ticks = 0;
    printf("setup timer interrupts\n");
}

void clock_set_next_interrupt(void){
    int64_t error_code = sbi_set_timer(get_time() + TIMEBASE);
    if(error_code == 0) return;
    else{
        //mark 输出错误代码 完善stdio
        panic("sbi set timer");
    }
}