#include "vm.h"
#include "memory.h"
#include "string.h"
#include "riscv.h"
#include "stdio.h"
#include "register.h"

pgtable_t kernel_pagetable;
extern char text_End[];         // 定义在linker.ld  
                                // 内核程序文本段结束地址，定义为可读可执行

extern char trampoline[];       // 定义在linker.ld  蹦床
/*
 * @brief 初始化内核虚拟内存，建立va到pa映射
*/
void init_kernelvm(){
    kernel_pagetable = (pgtable_t)alloc_page();
    if(kernel_pagetable == 0){
        panic("init_kernelvm failed");
    }
    memset(kernel_pagetable, 0, PGSIZE);

/*
    // uart registers
    kvmmap(kpgtbl, UART0, UART0, PGSIZE, PTE_R | PTE_W);

    // virtio mmio disk interface
    kvmmap(kpgtbl, VIRTIO0, VIRTIO0, PGSIZE, PTE_R | PTE_W);

    // PLIC
    kvmmap(kpgtbl, PLIC, PLIC, 0x400000, PTE_R | PTE_W);
*/
    // map kernel text executable and read-only.
    map_pages(kernel_pagetable, KERNEL_ENTRY, (uint64_t)text_End-KERNEL_ENTRY, KERNEL_ENTRY, PTE_R | PTE_X);
    // map kernel data and the physical RAM we'll make use of.
    // 分两段映射是因为两段权限不一样，text可读可执行，其他是可读可写
    map_pages(kernel_pagetable, (uint64_t)text_End, PYHSTOP-(uint64_t)text_End, (uint64_t)text_End, PTE_R | PTE_W);
    // map the trampoline for trap entry/exit to
    // the highest virtual address in the kernel.
    //kvmmap(kpgtbl, TRAMPOLINE, (uint64)trampoline, PGSIZE, PTE_R | PTE_X);
    map_pages(kernel_pagetable, TRAMPOLINE, PGSIZE, (uint64_t)trampoline, PTE_R | PTE_W);
    
    printf("kernel virtual memory finished!\n");
}

void init_hart_vm(){
    //printf("1\n");
    w_satp((((uint64_t)kernel_pagetable)>>12) | SATP_MODE);
    //printf("2\n");
    //刷新TLB
    sfence_vma();
    printf("init_hart_vm finished!\n");
}
/*
 * @brief 在pagetable中创建从va到pa的映射
 * @param pagetable 页表
 * @param va 虚拟地址   
 * @param size 大小 
 * @param pa 物理地址
 * @param perm pte的标志位   
 * @retval flase or true 
*/
bool map_pages(pgtable_t pagetable, uint64_t va, uint64_t size, uint64_t pa, int perm){
    if(size == 0)
        panic("map_pages");
    uint64_t tmp = PGROUNDDOWN(va);
    uint64_t last = PGROUNDDOWN(va+size-1);
    pte_t *pte;
    for(; tmp <= last; tmp+=PGSIZE,pa+=PGSIZE){
        if((pte = get_last_pgtable(pagetable, tmp, 1)) == 0){
            //panic("map_pages error");
            return false;
        }
        if(*pte & PTE_V){
            //panic("remap");
            return false;
        }
        *pte = PA2PTE(pa) | perm | PTE_V;
    }
    return true;
}

/*
 * @brief 将虚拟地址va起n页所对应的物理页解除在页表中的映射 
 * @note va必须页对齐
 * @param do_free 为1时释放对应的物理页
*/
bool free_map(pgtable_t pagetable, uint64_t va, uint64_t npages, int32_t do_free){
    pte_t *pte;
    if((va % PGSIZE) != 0)
        panic("uvmunmap: not aligned");
    for(uint64_t i = va; i < va + npages*PGSIZE; i+=PGSIZE){
        if((pte=get_last_pgtable(pagetable, i, 0)) == false)
            panic("free_map: get_last_pgtable error");
        if((*pte & PTE_V) == 0)
            panic("free_map: PTE error");
        if(do_free){
            uint64_t pa = PTE2PA(*pte);
            memfree(pa);
        }
        *pte = 0;
    }
    return true;
}

/*
 * @brief 从三级页表得到一级页表
 * @param pagetable:三级页表
 * @param va:虚拟地址 
 * @param alloc:=1时缺页时允许分配页
 * 
*/
pte_t* get_last_pgtable(pgtable_t pagetable, uint64_t va, int alloc){
    if(va >= MAXVA)    panic("next_pgtable: the virtual address is overflow");
    for(int level = 2; level > 0; level--){
        pte_t *pte = &pagetable[(va>>(PG_OFFSET+level*PPN_OFFSET)) & PGVA_MASK];
        if(*pte & PTE_V){
            pagetable = (pgtable_t)PTE2PA(*pte);
        }else{
            if(alloc == 0 || (pagetable = (pgtable_t)alloc_page()) == 0){
                return 0;
            }
            memset(pagetable, 0, PGSIZE);
            *pte = PA2PTE(pagetable) | PTE_V;
        }
    }

    return pagetable + ((va>>PG_OFFSET) & PGVA_MASK);
}

/*
 * @brief 将父进程的页表复制给子进程
*/
bool uvmcopy(pgtable_t old, pgtable_t new, uint64_t sz){
    pte_t *pte;
    uint64_t pa, i;
    uint32_t flags;
    char *mem;

    for(i = 0; i < sz; i += PGSIZE){
        if((pte = get_last_pgtable(old, i, 0)) == 0)
            panic("uvmcopy: pte should exist");
        if((*pte & PTE_V) == 0)
            panic("uvmcopy: page not present");
        pa = PTE2PA(*pte);
        flags = PTE_FLAGS(*pte);
        if((mem = alloc_page()) == 0)
            goto err;
        memcpy(mem, (char*)pa, PGSIZE);
        if(map_pages(new, i, PGSIZE, (uint64_t)mem, flags) != 0){
            memfree((uint64_t)mem);
            goto err;
        }
    }
    return 0;

    err:
    free_map(new, 0, i / PGSIZE, 1);
    return -1;
}